__all__ = (
    "ALLOWED_AROMATICITY_MODELS",
    "ALLOWED_CHARGE_MODELS",
    "ALLOWED_FRACTIONAL_BOND_ORDER_MODELS",
    "DEFAULT_AROMATICITY_MODEL",
    "DEFAULT_CHARGE_MODEL",
    "DEFAULT_FRACTIONAL_BOND_ORDER_MODEL",
)


ALLOWED_AROMATICITY_MODELS = ["OEAroModel_MDL"]
DEFAULT_AROMATICITY_MODEL = ALLOWED_AROMATICITY_MODELS[0]

ALLOWED_FRACTIONAL_BOND_ORDER_MODELS = ["Wiberg"]
DEFAULT_FRACTIONAL_BOND_ORDER_MODEL = ALLOWED_FRACTIONAL_BOND_ORDER_MODELS[0]

ALLOWED_CHARGE_MODELS = ["AM1-BCC"]
DEFAULT_CHARGE_MODEL = ALLOWED_CHARGE_MODELS[0]
